/*******************************************************************************
* ParTypes VCC library :: Protocol definition.
******************************************************************************/

#ifndef __Protocol_h__
#define __Protocol_h__ /* { */

// NOTE ON VCC SYNTAX: use only one ghost definition per _( section ) 
// or vcc will not eat it. 
// Several axioms can be in the same _(section) though.

// Data types.
_(ghost typedef \bool IntPred [\integer];)
_(ghost typedef \bool IntArrayPred [int*][\integer];)
_(ghost typedef \bool FloatArrayPred [float*] [\integer];)

_(
datatype Data {
  case intRefin         (IntPred);
  case intArrayRefin    (IntArrayPred);
  case floatArrayRefin  (FloatArrayPred);
}
)
_(
axiom \forall IntPred a,b;
  a == b ==> intRefin(a) == intRefin(b);

axiom \forall IntArrayPred a,b;
  a == b ==> intArrayRefin(a) == intArrayRefin(b);

axiom \forall FloatArrayPred a,b;
  a == b ==> floatArrayRefin(a) == floatArrayRefin(b);
)

_(ghost typedef struct _vcc_math_type_Protocol IntAbs[\integer])
_(ghost typedef struct _vcc_math_type_Protocol IntArrayAbs[int*][\integer])
_(ghost typedef struct _vcc_math_type_Protocol FloatArrayAbs[float*][\integer])

_( 
datatype Protocol 
{
case skip            ();
case size            (IntPred, IntAbs);
case seq             (Protocol, Protocol);
case choice          (\bool, Protocol, Protocol);	
case foreach         (\integer, IntAbs);
case intVal          (IntPred, IntAbs);
case intArrayVal     (IntArrayPred, IntArrayAbs);
case floatArrayVal   (FloatArrayPred, FloatArrayAbs);
case message         (\integer, \integer, Data);
case intBcast        (\integer, IntPred, IntAbs); 
case intArrayBcast   (\integer, IntArrayPred, IntArrayAbs); 
case floatArrayBcast (\integer, FloatArrayPred, FloatArrayAbs);
case scatter         (\integer, Data);
case gather          (\integer, Data);
case allgather       (Data);
case reduce          (\integer, MPI_Op, Data);
case allreduce       (MPI_Op, Data);
}
)
_(
axiom \forall IntAbs a,b; \integer n;
  a == b ==> foreach(n,a) == foreach(n,b);

axiom \forall Data a,b; \integer from, to;
  a == b ==> message(from, to, a) == message(from, to, b);

axiom \forall IntPred a1,b1; IntAbs a2,b2; 
  a1 == b1 && a2 == b2 ==> intVal(a1,a2) == intVal(b1,b2);

axiom \forall IntPred a1,b1; IntAbs a2,b2; \integer root;
  a1 == b1 && a2 == b2 ==> intBcast(root,a1,a2) == intBcast(root,b1,b2);

axiom \forall IntArrayPred a1,b1; IntArrayAbs a2,b2; 
  a1 == b1 && a2 == b2 ==> intArrayVal(a1,a2) == intArrayVal(b1,b2);

axiom \forall IntArrayPred a1,b1; IntArrayAbs a2,b2; \integer root;
  a1 == b1 && a2 == b2 ==> intArrayBcast(root,a1,a2) == intArrayBcast(root,b1,b2);

axiom \forall FloatArrayPred a1,b1; FloatArrayAbs a2,b2; 
  a1 == b1 && a2 == b2 ==> floatArrayVal(a1,a2) == floatArrayVal(b1,b2);

axiom \forall FloatArrayPred a1,b1; FloatArrayAbs a2,b2; \integer root;
  a1 == b1 && a2 == b2 ==> floatArrayBcast(root,a1,a2) == floatArrayBcast(root,b1,b2);

axiom \forall Data a,b; \integer root; MPI_Op op; 
  a == b ==> reduce(root, op, a) == reduce(root, op, b);

axiom \forall Data a,b; MPI_Op op; 
  a == b ==> allreduce(op, a) == allreduce(op, b);

)
// isSkip predicate()

_(pure \bool isSkip(Protocol p, \integer rank))
_(
axiom \forall Protocol p; \integer rank;
  isSkip(p,rank) <==> head(cons(p,rank)) == skip();
)

// Cons (head-tail pair) datum
_(datatype Cons {
   case pair(Protocol, Protocol);
})
_(ghost _(pure) Protocol head(Cons c);)
_(ghost _(pure) Protocol tail(Cons c);)
_(ghost _(pure) Cons cons(Protocol p, \integer rank);)

_(
axiom \forall Protocol p1, p2; head(pair(p1,p2)) == p1;
axiom \forall Protocol p1, p2; tail(pair(p1,p2)) == p2;
axiom \forall Protocol a1,a2,b1,b2; 
  a1 == b1 && a2 == b2 ==> pair(a1,a2) == pair(b1,b2);
)

_(

/* 1. Definitions for the seq constructor

We need three cases: seq/seq, seq/skip, seq/others
*/

// seq/seq
axiom \forall Protocol p1,p2,p3; \integer rank;
    cons(seq(seq(p1,p2),p3),rank) == cons(seq(p1,seq(p2,p3)),rank);

// seq/skip
axiom \forall Protocol p; \integer rank;
  cons(seq(skip(),p),rank) == cons(p,rank);

// seq/skip in left-most leaf
axiom \forall Protocol p1,p2; \integer rank;
  cons(p1,rank) == pair(skip(),skip()) ==>
    cons(seq(p1,p2),rank) == cons(p2,rank);

// seq/skip all-other cases
axiom \forall Protocol p1,p2,h; \integer rank;
  cons(p1,rank) == pair(h,skip()) && h != skip() ==>
    cons(seq(p1,p2),rank) == pair(h,p2);


/* 2. Definitions for all non-seq constructors

We need one axiom for each non-seq type constructor, except for
messages where we need two: when the message is equivalent to skip,
and when it is not.
*/

axiom \forall \integer rank;
  cons(skip(),rank) == pair(skip(),skip()); 

axiom \forall \integer from,to; Data d; \integer rank;
  (rank == from || rank == to)  
    ==> cons(message(from,to,d),rank) == pair(message(from,to,d),skip()); 

axiom \forall \integer from,to; Data d; \integer rank;
  (rank != from && rank != to)  
    ==> cons(message(from,to,d),rank) == pair(skip(),skip());


axiom \forall IntPred pred; IntAbs abs; \integer rank;
  cons(size(pred,abs),rank) == pair(size(pred,abs),skip()); 

axiom \forall \bool b; Protocol p1,p2; \integer rank; 
  cons(choice(b,p1,p2),rank) == pair(choice(b,p1,p2),skip()); 

axiom \forall \integer n; IntAbs a; \integer rank; 
  cons(foreach(n,a),rank) == pair(foreach(n,a),skip()); 

axiom \forall IntPred pred; IntAbs abs; \integer rank; 
  cons(intVal(pred,abs),rank) == pair(intVal(pred,abs),skip()); 

axiom \forall IntArrayPred pred; IntArrayAbs abs; \integer rank; 
  cons(intArrayVal(pred,abs),rank) == pair(intArrayVal(pred,abs),skip()); 

axiom \forall FloatArrayPred pred; FloatArrayAbs abs; \integer rank; 
  cons(floatArrayVal(pred,abs),rank) == pair(floatArrayVal(pred,abs),skip()); 

axiom \forall IntPred pred; IntAbs abs; \integer rank, root; 
  cons(intBcast(root,pred,abs),rank) == pair(intBcast(root,pred,abs),skip()); 

axiom \forall IntArrayPred pred; IntArrayAbs abs; \integer rank, root; 
  cons(intArrayBcast(root,pred,abs),rank) == pair(intArrayBcast(root,pred,abs),skip()); 

axiom \forall FloatArrayPred pred; FloatArrayAbs abs; \integer rank, root; 
  cons(floatArrayBcast(root,pred,abs),rank) == pair(floatArrayBcast(root,pred,abs),skip()); 

axiom \forall Data d; \integer rank; 
  cons(allgather(d),rank) == pair(allgather(d),skip());

axiom \forall Data d; \integer rank, root; 
  cons(gather(root,d),rank) == pair(gather(root,d),skip());

axiom \forall Data d; \integer rank, root; 
  cons(scatter(root,d),rank) == pair(scatter(root,d), skip());

axiom \forall Data d; \integer rank, root; MPI_Op op; 
  cons(reduce(root,op,d),rank) == pair(reduce(root,op,d),skip());

axiom \forall Data d; \integer rank; MPI_Op op; 
  cons(allreduce(op,d),rank) == pair(allreduce(op,d),skip());
)


#endif /* } */
